#!/usr/bin/env python

import os
import numpy as np
import math
from math import pi
import matplotlib.pyplot as plt
import gym
import gym.spaces
import rospy
import roslib.packages
from gazebo_msgs.srv import SpawnModel, DeleteModel
from geometry_msgs.msg import Pose
from nav_msgs.msg import Odometry
from tf.transformations import euler_from_quaternion
from act_drive_17 import actDrive


class Env(gym.core.Env):
    def __init__(self, state_size, action_size, stage):
        self.action_size = action_size
        self.init_start = [-2.0, -0.8, 0.0]
        self.start = self.init_start
        if stage == 1:
            self.init_goal = [-0.5, -0.5]
        elif stage == 2:
            self.init_goal = [0.5, -0.5]
        else:
            self.init_goal = [1.7, -1.0]
        self.stepCnt = 0
        self.totalSteps = 0
        self.stepRewSum = 0.
        self.totalGoals = 0
        self.totalOuts = 0
        self.initGoal = True
        self.goForward = True
        self.get_goalbox = False
        self.position = Pose().position
        self.heading = 0
        self.goal_x = 0
        self.goal_y = 0
        self.goal_pos = Pose()
        self.yaw = 0.
        self.keep_turn = 0
        self.last_get_dir = rospy.Time()
        self.last_goal = rospy.Time()
        self.goalDurs = [0.]
        gmPkgPath = roslib.packages.get_pkg_dir('turtlebot3_gazebo')
        goalModelPath = os.path.join(
            gmPkgPath, './models/turtlebot3_square/goal_box/model.sdf')
        self.goalModel = open(goalModelPath, 'r').read()
        # self.pub_cmd_vel = rospy.Publisher('cmd_vel', Twist, queue_size=5)
        self.sub_odom = rospy.Subscriber('odom', Odometry, self.getOdometry)
        # 行動空間を設定
        self.action_space = gym.spaces.Discrete(action_size)
        # 観測空間を設定: 64行目のforループによる変換で最小値は0、最大値は3.5
        # Pythonでは要素数が1つのタプルは(R,)と表記するので１次元配列のshapeは(R,)となる
        self.observation_space = gym.spaces.Box(
            low=0., high=3.5, shape=(state_size,))
        # tb3_0のスキャンと速度制御を行うactDriveクラスのインスタンスを生成
        self.drive = actDrive('/cmd_vel')

    def respawnModel(self, modelName):
        rospy.wait_for_service('gazebo/spawn_sdf_model')
        spawn_model_prox = rospy.ServiceProxy(
            'gazebo/spawn_sdf_model', SpawnModel)
        spawn_model_prox(modelName, self.goalModel,
                         'robotos_name_space',
                         self.goal_pos, 'world')

    def deleteModel(self, modelName):
        rospy.wait_for_service('gazebo/delete_model')
        del_model_prox = rospy.ServiceProxy(
            'gazebo/delete_model', DeleteModel)
        del_model_prox(modelName)

    def getGoalDistace(self):
        goal_distance = round(math.hypot(self.goal_x - self.position.x,
                              self.goal_y - self.position.y), 2)
        return goal_distance

    def getOdometry(self, odom):
        self.position = odom.pose.pose.position
        orientation = odom.pose.pose.orientation
        orientation_list = [orientation.x, orientation.y,
                            orientation.z, orientation.w]
        _, _, self.yaw = euler_from_quaternion(orientation_list)
        goal_angle = math.atan2(self.goal_y - self.position.y,
                                self.goal_x - self.position.x)
        # headingはゴールの方位と自身のyaw角の差分(-π〜πのラジアン)
        heading = goal_angle - self.yaw
        if heading > pi:
            heading -= 2 * pi
        elif heading < -pi:
            heading += 2 * pi
        self.heading = round(heading, 2)

    def getState(self):
        while self.drive.scan_range == []:
            rospy.sleep(0.01)
        scan_range = self.drive.scan_range
        # print(f"scan_range: {scan_range}")
        heading = self.heading
        current_distance = round(math.hypot(
            self.goal_x - self.position.x, self.goal_y - self.position.y), 2)
        obstacle_min_range = round(min(scan_range), 2)
        obstacle_angle = np.argmin(scan_range)
        state = scan_range + [heading, current_distance,
                              obstacle_min_range, obstacle_angle]
        done = False
        # 250ステップになるとdoneをTrueにする
        if self.stepCnt >= 250:
            done = True
        # ゴールするとdoneをTrueにする
        if current_distance < 0.2:
            self.get_goalbox = True
            done = True
        return state, done

    def setReward(self, state, done, action):
        yaw_reward = []
        current_distance = state[-3]
        heading = state[-4]
        if self.get_goalbox:
            self.totalGoals += 1
            print(f"Goal!!  {self.totalGoals} times")
            reward = 200.
            # self.pub_cmd_vel.publish(Twist())
            self.drive.pubInitVel()
        elif done:
            self.totalOuts += 1
            print(f"Timeout!!  {self.totalOuts} times")
            reward = -150.
            # self.pub_cmd_vel.publish(Twist())
            self.drive.pubInitVel()
        else:
            # 各actionに対応するyaw_rewardリストを生成する
            for i in range(5):
                tr = 1. - 2 * math.fabs((heading / pi
                                        + (i - 2) / 8 + 1) % 2 - 1)
                yaw_reward.append(tr)
            distance_rate = 2 ** (current_distance / self.goal_distance)
            reward = ((round(yaw_reward[action] * 5, 2)) * distance_rate)
        return reward

    def plot_durs(self):
        plt.close("all")
        plt.rcParams['font.size'] = 14
        plt.rcParams['xtick.direction'] = 'in'
        plt.rcParams['ytick.direction'] = 'in'
        fig, ax = plt.subplots(1, 1, figsize=(8.0, 4.5))
        totalTime = sum(self.goalDurs)
        meanDur = round(totalTime / self.totalGoals, 1) \
            if self.totalGoals != 0 else 0
        ax.set_title(f"Goals: {self.totalGoals}     "
                     f"Mean Duration: {meanDur} s")
        ax.set_xlabel('Number of Goals', fontsize=16)
        ax.set_ylabel('Duration (sec)', fontsize=16)
        ax.set_position([0.14, 0.15, 0.8, 0.75])
        # ax.plot(self.episRewLog)
        ax.plot(self.goalDurs)
        ax.set_xlim(left=1)
        ax.set_ylim(bottom=0.)
        # 最後のプロットをファイル保存するためにpltをインスタンス変数として保存
        self.plt = plt
        plt.pause(0.1)

    def step(self, action):
        state, done = self.getState()
        keep = self.drive.pubVel(action)
        reward = self.setReward(state, done, action)
        # ステップ毎のrewardを今までの合計に加算し、ステップカウンタをカウントアップ
        self.stepRewSum += reward
        self.stepCnt += 1
        self.totalSteps += 1
        # ステップ数と報酬に加え、進路選択状況を見やすく表示するkeepを出力
        print(f"Step: {self.stepCnt:3d}, ",
              f"Sum: {self.totalSteps:4d}, ",
              f"Rew: {reward:>7.2f}, ",
              f"Sum: {self.stepRewSum:>7.2f}, ",
              f"keep: {keep}")
        # "matplotlibの応答がありません"という警告を回避するため20ステップ毎にグラフを更新
        if self.stepCnt % 20 == 1:
            self.plot_durs()
        # 次のobservation、reward、done、追加情報(特にないので空dict)を順番に返す
        return np.asarray(state), reward, done, {}

    def reset(self):
        # エピソードが終了したら平均報酬をログリストに記録し、ステップ内カウントをリセット
        if self.stepCnt != 0:
            meanRew = self.stepRewSum / self.stepCnt
            print(f"Episode Mean Reward : {meanRew}")
            # self.episRewLog.append(meanRew)
            self.stepCnt = 0
            self.stepRewSum = 0.
        if self.initGoal or self.get_goalbox:
            if self.initGoal:
                self.initGoal = False
                self.last_goal = rospy.Time.now()
            else:
                self.get_goalbox = False
                self.goForward = not self.goForward
                self.start = [self.position.x, self.position.y, self.yaw]
                if self.goForward:
                    self.deleteModel('goal2')
                else:
                    self.deleteModel('goal1')
                # ゴールに要したfloat型の秒数をリストgoalDursに追記
                self.goalDurs.append(
                    (rospy.Time.now() - self.last_goal).to_sec())
                self.last_goal = rospy.Time.now()
            if self.goForward:
                self.goal_pos.position.x = self.init_goal[0]
                self.goal_pos.position.y = self.init_goal[1]
                self.respawnModel('goal1')
            else:
                self.goal_pos.position.x = self.init_start[0]
                self.goal_pos.position.y = self.init_start[1]
                self.respawnModel('goal2')
            self.goal_x = self.goal_pos.position.x
            self.goal_y = self.goal_pos.position.y
        state, done = self.getState()
        self.goal_distance = self.getGoalDistace()
        return np.asarray(state)
