#!/usr/bin/env python

import rospy
import shlex
from psutil import Popen
from gazebo_msgs.srv import DeleteModel
import gym
import gym.spaces
import numpy as np
import math
from math import pi
import matplotlib.pyplot as plt
from geometry_msgs.msg import Twist, Pose
from sensor_msgs.msg import LaserScan
from nav_msgs.msg import Odometry
from tf.transformations import euler_from_quaternion
from respawnGoal import Respawn


class Env(gym.core.Env):
    def __init__(self, state_size, action_size):
        # ここで初めてtb3_0をspawnする
        self.node_process = Popen(
            shlex.split("roslaunch turtlebot3_dqn tb3_0.launch")
        )
        self.goal_x = 0
        self.goal_y = 0
        self.heading = 0
        self.action_size = action_size
        self.stepCnt = 0
        self.totalSteps = 0
        self.stepRewSum = 0.
        self.episRewLog = [0.]
        self.totalGoals = 0
        self.totalColls = 0
        self.initGoal = True
        self.get_goalbox = False
        self.position = Pose().position
        self.pub_cmd_vel = rospy.Publisher('cmd_vel', Twist, queue_size=5)
        self.sub_odom = rospy.Subscriber('odom', Odometry, self.getOdometry)
        self.respawn_goal = Respawn()
        # 行動空間を設定
        self.action_space = gym.spaces.Discrete(action_size)
        # 観測空間を設定: 64行目のforループによる変換で最小値は0、最大値は3.5
        # Pythonでは要素数が1つのタプルは(R,)と表記するので１次元配列のshapeは(R,)となる
        self.observation_space = gym.spaces.Box(
            low=0., high=3.5, shape=(state_size,))

    def deleteModel(self, modelName):
        rospy.wait_for_service('gazebo/delete_model')
        del_model_prox = rospy.ServiceProxy(
            'gazebo/delete_model', DeleteModel)
        del_model_prox(modelName)

    def getGoalDistace(self):
        goal_distance = round(math.hypot(self.goal_x - self.position.x,
                              self.goal_y - self.position.y), 2)
        return goal_distance

    def getOdometry(self, odom):
        self.position = odom.pose.pose.position
        orientation = odom.pose.pose.orientation
        orientation_list = [orientation.x, orientation.y,
                            orientation.z, orientation.w]
        _, _, yaw = euler_from_quaternion(orientation_list)
        goal_angle = math.atan2(self.goal_y - self.position.y,
                                self.goal_x - self.position.x)
        # headingはゴールの方位と自身のyaw角の差分(-π〜πのラジアン)
        heading = goal_angle - yaw
        if heading > pi:
            heading -= 2 * pi
        elif heading < -pi:
            heading += 2 * pi
        self.heading = round(heading, 2)

    def getState(self):
        scan_range = []
        heading = self.heading
        min_range = 0.17
        done = False
        scan_data = None
        while scan_data is None:
            try:
                scan_data = rospy.wait_for_message(
                    'scan', LaserScan, timeout=5)
            except BaseException:
                pass
        scan_data2 = None
        while scan_data2 is None:
            try:
                scan_data2 = rospy.wait_for_message(
                    'scan2', LaserScan, timeout=5)
            except BaseException:
                pass
        scan_both = [scan_data, scan_data2]
        for i in range(len(scan_data.ranges)):
            if 5 < i < 18:
                j = 0
            else:
                j = 1
            if scan_both[j].ranges[i] == float('Inf'):
                scan_range.append(3.5)
            elif np.isnan(scan_both[j].ranges[i]):
                scan_range.append(0)
            else:
                scan_range.append(scan_both[j].ranges[i])
        obstacle_min_range = round(min(scan_range), 2)
        obstacle_angle = np.argmin(scan_range)
        if min_range > min(scan_range) > 0:
            done = True
        current_distance = round(math.hypot(
            self.goal_x - self.position.x, self.goal_y - self.position.y), 2)
        if current_distance < 0.2:
            self.get_goalbox = True
            done = True
        return scan_range + [heading, current_distance,
                             obstacle_min_range, obstacle_angle], done

    def setReward(self, state, done, action):
        yaw_reward = []
        current_distance = state[-3]
        heading = state[-4]
        if self.get_goalbox:
            self.totalGoals += 1
            print(f"Goal!!  {self.totalGoals} times")
            reward = 200.
            self.pub_cmd_vel.publish(Twist())
        elif done:
            self.totalColls += 1
            print(f"Collision!!  {self.totalColls} times")
            reward = -150.
            self.pub_cmd_vel.publish(Twist())
        else:
            # 各actionに対応するyaw_rewardリストを生成する(headingのズレに応じて0〜-10)
            for i in range(5):
                tr = -10. * math.fabs((heading / pi
                                      + (i - 2) / 8 + 1) % 2 - 1)
                yaw_reward.append(tr)
            # dist_rewardはゴールまでの距離に応じて0〜-5(スタート時より遠ざかれば<-5に)
            dist_reward = -5. * (current_distance / self.goal_distance)
            print(f"dist_rew: {dist_reward:.1f},  "
                  f"yaw_rew: {yaw_reward[action]:.1f}")
            # エピソードが長引くほど負の報酬が積算される
            reward = yaw_reward[action] + dist_reward
        return reward

    def plot_reward(self):
        plt.close("all")
        plt.rcParams['font.size'] = 14
        plt.rcParams['xtick.direction'] = 'in'
        plt.rcParams['ytick.direction'] = 'in'
        fig, ax = plt.subplots(1, 1, figsize=(8.0, 4.5))
        ax.set_title(f"Goals: {self.totalGoals}     "
                     f"Collisions: {self.totalColls}")
        ax.set_xlabel('Episodes', fontsize=16)
        ax.set_ylabel('Mean Reward', fontsize=16)
        ax.set_position([0.14, 0.15, 0.8, 0.75])
        ax.plot(self.episRewLog)
        # 最後のプロットをファイル保存するためにpltをインスタンス変数として保存
        self.plt = plt
        plt.pause(0.1)

    def step(self, action):
        max_linear_vel = 0.15
        max_angular_vel = 1.5
        vel_cmd = Twist()
        # action(0〜4)によって並進速度を設定
        vel_cmd.linear.x = max_linear_vel * [1., 1., 1., 1., 1.][action]
        # action(0〜4)によって角速度を設定
        vel_cmd.angular.z = max_angular_vel * [1., 0.5, 0., -0.5, -1.][action]
        self.pub_cmd_vel.publish(vel_cmd)
        state, done = self.getState()
        reward = self.setReward(state, done, action)
        # ステップ毎のrewardを今までの合計に加算し、ステップカウンタをカウントアップする
        self.stepRewSum += reward
        self.stepCnt += 1
        self.totalSteps += 1
        print(f"Step: {self.stepCnt:3d}, ",
              f"Sum: {self.totalSteps:4d}, ",
              f"Rew: {reward:>7.2f}, ",
              f"Sum: {self.stepRewSum:>7.2f}, ",
              f"Scan: min {state[-2]:.2f}  at {state[-1]:>2d}")
        # print([f"{x:.2f}" for x in state[0:24]])
        # "matplotlibの応答がありません"という警告を回避するため20ステップ毎にグラフを更新
        if self.stepCnt % 20 == 1:
            self.plot_reward()
        # 次のobservation、reward、done、追加情報(特にないので空dict)を順番に返す
        return np.asarray(state), reward, done, {}

    def reset(self):
        # エピソードが終了したら平均報酬をログリストに記録し、ステップ内カウントをリセット
        if self.stepCnt != 0:
            meanRew = self.stepRewSum / self.stepCnt
            print(f"Episode Mean Reward : {meanRew}")
            self.episRewLog.append(meanRew)
            self.stepCnt = 0
            self.stepRewSum = 0.
            # "matplotlibの応答がありません"という警告を回避するため20ステップ毎にグラフを更新
            # self.plot_reward()
        if self.initGoal:
            self.initGoal = False
            self.goal_x, self.goal_y = self.respawn_goal.getPosition()
        else:
            if self.get_goalbox:
                self.get_goalbox = False
                self.goal_x, self.goal_y = self.respawn_goal.getPosition(
                    True, delete=True)
            # tb3_0.launchで起動されたノードを終了し再起動する
            self.deleteModel('tb3_0')
            self.node_process.terminate()
            self.node_process = Popen(
                shlex.split("roslaunch turtlebot3_dqn tb3_0.launch")
            )
            # odomデータによる更新が遅れてself.getGoalDistace()が間違った値を返すのを防止
            self.position.x = 0.
            self.position.y = 0.
        self.goal_distance = self.getGoalDistace()
        state, done = self.getState()
        return np.asarray(state)
