/*******************************************************************************
* Copyright 2016 ROBOTIS CO., LTD.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*******************************************************************************/

/* Authors: Taehun Lim (Darby) */

#include "turtlebot3_gazebo/turtlebot3_drive.h"

Turtlebot3Drive::Turtlebot3Drive()
  : nh_priv_("~")
{
  //Init gazebo ros turtlebot3 node
  ROS_INFO("TurtleBot3 Simulation Node Init");
  auto ret = init();
  ROS_ASSERT(ret);
}

Turtlebot3Drive::~Turtlebot3Drive()
{
  updatecommandVelocity(0.0, 0.0);
  ros::shutdown();
}

/*******************************************************************************
* Init function
*******************************************************************************/
bool Turtlebot3Drive::init()
{
  // initialize ROS parameter
  std::string cmd_vel_topic_name = nh_.param<std::string>("cmd_vel_topic_name", "");

  // initialize variables
  escape_range_ = 3.0 * DEG2RAD;            // この角度(rad)だけ旋回したら進路を決定し直す
  turn_timeout_ = 5.0;                      // この時間内に上記旋回ができなかったら旋回方向を反転する
  keep_turn_ = 0;                           // 左旋回を開始したら1、右旋回を開始したら-1

  check_front_dist_ = 0.7;
  check_forward_dist_ = 0.5;
  check_side_dist_ = 0.3;

  tb3_pose_ = 0.0;                          // 座標軸に対する現時点のAMRの向き(rad)
  prev_tb3_pose_ = 0.0;                     // 旋回を開始する前のAMRの向き(rad)

  // initialize publishers
  cmd_vel_pub_   = nh_.advertise<geometry_msgs::Twist>(cmd_vel_topic_name, 10);

  // initialize subscribers
  laser_scan_sub_  = nh_.subscribe("scan2", 10, &Turtlebot3Drive::laserScanMsgCallBack, this);
  odom_sub_ = nh_.subscribe("odom", 10, &Turtlebot3Drive::odomMsgCallBack, this);

  return true;
}

void Turtlebot3Drive::odomMsgCallBack(const nav_msgs::Odometry::ConstPtr &msg)
{
  double siny = 2.0 * (msg->pose.pose.orientation.w * msg->pose.pose.orientation.z + msg->pose.pose.orientation.x * msg->pose.pose.orientation.y);
  double cosy = 1.0 - 2.0 * (msg->pose.pose.orientation.y * msg->pose.pose.orientation.y + msg->pose.pose.orientation.z * msg->pose.pose.orientation.z);

  tb3_pose_ = atan2(siny, cosy);
}

void Turtlebot3Drive::laserScanMsgCallBack(const sensor_msgs::LaserScan::ConstPtr &msg)
{
  // 1°単位のスキャンで何番目のデータかを示す指数
  uint16_t scan_angle[6] = {10, 350, 20, 340, 50, 310};
  // 左右の鋭い角度範囲内のスキャン距離の最小値を取得
  scan_front_ = msg->range_max;
  for(int num = 0; num < scan_angle[0]; num++) {
    if(msg->ranges.at(num) < scan_front_) {
      scan_front_ = msg->ranges.at(num);
    }
  }
  for(int num = scan_angle[1]; num < 360; num++) {
    if(msg->ranges.at(num) < scan_front_) {
      scan_front_ = msg->ranges.at(num);
    }
  }
  // 左右の狭い角度範囲内のスキャン距離の最小値を取得
  scan_forward_ = msg->range_max;
  for(int num = 0; num < scan_angle[2]; num++) {
    if(msg->ranges.at(num) < scan_forward_) {
      scan_forward_ = msg->ranges.at(num);
      min_is_left_ = true;
    }
  }
  for(int num = scan_angle[3]; num < 360; num++) {
    if(msg->ranges.at(num) < scan_forward_) {
      scan_forward_ = msg->ranges.at(num);
      min_is_left_ = false;
    }
  }
  // 左右の広い角度範囲内のスキャン距離の最小値を取得
  scan_side_ = msg->range_max;
  for(int num = 0; num < scan_angle[4]; num++) {
    if(msg->ranges.at(num) < scan_side_) {
      scan_side_ = msg->ranges.at(num);
      if(scan_side_ < check_side_dist_) min_is_left_ = true;
    }
  }
  for(int num = scan_angle[5]; num < 360; num++) {
    if(msg->ranges.at(num) < scan_side_) {
      scan_side_ = msg->ranges.at(num);
      if(scan_side_ < check_side_dist_) min_is_left_ = false;
    }
  }
}

void Turtlebot3Drive::updatecommandVelocity(double linear, double angular)
{
  geometry_msgs::Twist cmd_vel;

  cmd_vel.linear.x  = linear;
  cmd_vel.angular.z = angular;

  cmd_vel_pub_.publish(cmd_vel);
}

/*******************************************************************************
* Control Loop function
*******************************************************************************/
bool Turtlebot3Drive::controlLoop()
{
  static uint8_t turtlebot3_state_num = 0;

//  switch(turtlebot3_state_num) {
//    case GET_TB3_DIRECTION:
//      last_get_direction_ = ros::Time::now();
//      if(scan_front_ > check_front_dist_
//          && scan_forward_ > check_forward_dist_
//          && scan_side_ > check_side_dist_) {
//        turtlebot3_state_num = TB3_DRIVE_FORWARD;
//        keep_turn_ = 0;
//      } else if(keep_turn_ == 1) {
//        prev_tb3_pose_ = tb3_pose_;
//        turtlebot3_state_num = TB3_LEFT_TURN;
//      } else if(keep_turn_ == -1) {
//        prev_tb3_pose_ = tb3_pose_;
//        turtlebot3_state_num = TB3_RIGHT_TURN;
//      } else if(min_is_left_) {
//        prev_tb3_pose_ = tb3_pose_;
//        turtlebot3_state_num = TB3_RIGHT_TURN;
//        keep_turn_ = -1;
//      } else {
//         prev_tb3_pose_ = tb3_pose_;
//         turtlebot3_state_num = TB3_LEFT_TURN;
//         keep_turn_ = 1;
//      }
//      break;

//    case TB3_DRIVE_FORWARD:
//      updatecommandVelocity(LINEAR_VELOCITY, 0.0);
//      turtlebot3_state_num = GET_TB3_DIRECTION;
//      break;

//    case TB3_RIGHT_TURN:
//      if(last_get_direction_ + ros::Duration(turn_timeout_) < ros::Time::now()) {
//        keep_turn_ = 1;
//      } else {
//        updatecommandVelocity(0.0, -1 * ANGULAR_VELOCITY);
//      }
//      turtlebot3_state_num = GET_TB3_DIRECTION;
//      break;

//    case TB3_LEFT_TURN:
//      if(last_get_direction_ + ros::Duration(turn_timeout_) < ros::Time::now()) {
//        keep_turn_ = -1;
//      } else {
//        updatecommandVelocity(0.0, ANGULAR_VELOCITY);
//      }
//      turtlebot3_state_num = GET_TB3_DIRECTION;
//      break;

//    default:
//      turtlebot3_state_num = GET_TB3_DIRECTION;
//      break;
//  }
  if(scan_front_ > check_front_dist_
      && scan_forward_ > check_forward_dist_
      && scan_side_ > check_side_dist_) {
    keep_turn_ = 0;
    updatecommandVelocity(LINEAR_VELOCITY, 0.0);
  } else if(keep_turn_ == 0) {
    last_get_direction_ = ros::Time::now();
    if(min_is_left_) {
      keep_turn_ = -1;
    } else {
      keep_turn_ = 1;
    }
  }
  if(keep_turn_ != 0) {
    if(last_get_direction_ + ros::Duration(turn_timeout_) < ros::Time::now()) {
      last_get_direction_ = ros::Time::now();
      keep_turn_ *= -1;
    } else {
      updatecommandVelocity(0.0, keep_turn_ * ANGULAR_VELOCITY);
    }
  }
  return true;
}

/*******************************************************************************
* Main function
*******************************************************************************/
int main(int argc, char* argv[])
{
  ros::init(argc, argv, "turtlebot3_drive");
  Turtlebot3Drive turtlebot3_drive;

  ros::Rate loop_rate(30);

  while (ros::ok())
  {
    turtlebot3_drive.controlLoop();
    ros::spinOnce();
    loop_rate.sleep();
  }

  return 0;
}
